# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Request views module."""

from .api import create_request_events_bp, create_requests_bp
from .ui import create_ui_blueprint

__all__ = (
    'create_ui_blueprint',
    'create_requests_bp',
    'create_request_events_bp',
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Request views decorators module."""

from functools import wraps

from flask import g

from invenio_requests.proxies import current_requests


def service():
    """Get the requests service."""
    return current_requests.requests_service


def pass_request(f):
    """Retrieve request record to the view."""

    @wraps(f)
    def view(**kwargs):
        """Decorated view."""
        pid_value = kwargs['request_pid_value']
        request = service().read(
            id_=pid_value, identity=g.identity
        )
        kwargs['request'] = request
        return f(**kwargs)

    return view



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Related record systemfield."""

from invenio_records.dictutils import dict_lookup, parse_lookup_key
from invenio_records.systemfields import SystemField


class AttrProxy:
    """Attribute proxy.

    The main purpose of this proxy, is to use data stored on the current record
    if available, instead of quering the database for the related record.
    """

    def __init__(self, record_cls, record, data, attrs=None):
        """Initialize the attribute proxy."""
        self._data = data
        self._attrs = attrs or []
        self._id = record.id if record else data['id']
        self._record = record
        self._record_shim = None
        self._record_cls = record_cls

    def get_object(self):
        """Get the underlying record."""
        if self._record is None:
            self._record = self._record_cls.get_record(self._id)
        return self._record

    def get_object_shim(self):
        """Get a record shim.

        The object shim is used for accessing attributes where we want the
        the proxied record's system field to be invoked. E.g. ``request.type``.
        """
        if self._record_shim is None:
            self._record_shim = self._record_cls(self._data)
        return self._record_shim

    def __getattr__(self, attr):
        """Attribute access."""
        if self._record is None:
            if attr == 'id':
                return self._id
            if attr in self._attrs:
                shim = self.get_object_shim()
                return getattr(shim, attr)
            self._record = self._record_cls.get_record(self._id)
        return getattr(self._record, attr)

    def __getitem__(self, attr):
        """Item access."""
        if self._record is None:
            if attr in self._attrs or attr == 'id':
                return self._data[attr]
            self._record = self._record_cls.get_record(self._id)
        return self._record[attr]


class RelatedRecord(SystemField):
    """Systemfield for managing a single related record.

    The field will save the 'id' and potentially more attribtues from the
    related record. When accessed, it will only load the related record if
    the attribute was not dumped inside the current record.

    CAUTION: This field should be used with care. It does not store the version
    counter from the related record, and thus you cannot detect stale data.
    """

    def __init__(self, record_cls, *args, keys=None, attrs=None, **kwargs):
        """Initialize field."""
        self._record_cls = record_cls
        # Dump just 'id' by default.
        self._dump_keys = keys or []
        self._dump_attrs = attrs or []
        super().__init__(*args, **kwargs)

    @property
    def _proxy_attrs(self):
        return set(self._dump_keys + self._dump_attrs)

    #
    # Life-cycle hooks
    #
    def _dump(self, proxy):
        """Dump given attributes from a proxy.

        Note, only top-level keys are supported.
        """
        data = {'id': str(proxy.id)}
        for k in self._dump_attrs:
            data[k] = getattr(proxy, k)
        for k in self._dump_keys:
            data[k] = proxy[k]

        # Add a version counter "@v" used for optimistic
        # concurrency control. It allows to search for all
        # outdated records and reindex them.
        data['@v'] = f'{proxy.id}::{proxy.revision_id}'
        return data

    def pre_commit(self, record):
        """Called before a record is committed."""
        proxy = getattr(record, self.attr_name)
        if proxy is not None:
            self.set_dictkey(
                record,
                self._dump(proxy),
                create_if_missing=True
            )

    #
    # Helpers
    #
    def _unset_cache(self, record):
        """Unset an object on the instance's cache."""
        if hasattr(record, '_obj_cache'):
            record._obj_cache.pop(self.attr_name, None)

    def del_value(self, record):
        """Delete the record relation."""
        # Unset cache if None.
        self._unset_cache(record)
        try:
            keys = parse_lookup_key(self.key)
            parent = dict_lookup(record, keys, parent=True)
            parent.pop(keys[-1], None)
        except KeyError:
            pass
        return

    def set_value(self, record, record_or_id):
        """Set the record (by id, record or proxy)."""
        # Unset cache if None.
        if record_or_id is None:
            self.del_value(record)
            return

        # Set value
        if isinstance(record_or_id, str):
            proxy = AttrProxy(
                self._record_cls,
                None,
                {'id': record_or_id},
                attrs=self._proxy_attrs
            )
        elif isinstance(record_or_id, self._record_cls):
            proxy = AttrProxy(
                self._record_cls,
                record_or_id,
                None,
                attrs=self._proxy_attrs
            )
        elif isinstance(record_or_id, AttrProxy):
            proxy = record_or_id
        else:
            raise ValueError("Invalid value.")
        self._set_cache(record, proxy)

    def get_value(self, record):
        """Get the record type."""
        proxy = self._get_cache(record)
        if proxy is not None:
            return proxy

        data = self.get_dictkey(record)
        if data is not None:
            proxy = AttrProxy(
                self._record_cls,
                None,
                data,
                attrs=self._proxy_attrs
            )
            self._set_cache(record, proxy)
        else:
            proxy = None

        return proxy

    #
    # Data descriptor methods (i.e. attribute access)
    #
    def __get__(self, record, owner=None):
        """Get the persistent identifier."""
        if record is None:
            return self
        return self.get_value(record)

    def __set__(self, record, pid):
        """Set persistent identifier on record."""
        self.set_value(record, pid)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Request permissions."""

import operator
from functools import reduce
from itertools import chain

from elasticsearch_dsl import Q
from invenio_records_permissions.generators import Generator
from invenio_records_resources.references import EntityGrant


class Status(Generator):
    """Generator to validate needs only for a given request status."""

    def __init__(self, statuses, generators):
        """Initialize need entity generator."""
        self._statuses = statuses
        self._generators = generators or []

    def needs(self, request=None, **kwargs):
        """Needs if status is in one of the provided ones."""
        if request.status in self._statuses:
            needs = [
                g.needs(request=request, **kwargs) for g in self._generators
            ]
            return set(chain.from_iterable(needs))
        return []

    def query_filter(self, **kwargs):
        """Query filters for the current identity."""
        queries = [g.query_filter(**kwargs) for g in self._generators]
        queries = [q for q in queries if q]
        queries = reduce(operator.or_, queries) if queries else None

        if queries:
            return Q("terms", **{"status": self._statuses}) & queries
        return None


class EntityNeedsGenerator(Generator):
    """Allows the creator of the request."""

    entity_field = None
    grants_field = 'grants'

    def __init__(self):
        """Initialize need entity generator."""
        assert self.entity_field is not None, "Subclass must define entity_field."

    def needs(self, request=None, **kwargs):
        """Needs for the given entity reference."""
        entity = getattr(request, self.entity_field)
        return request.type.entity_needs(entity)

    def query_filter(self, identity=None, **kwargs):
        """Query filters for the current identity."""
        grants = []
        for need in identity.provides:
            grants.append(EntityGrant(self.entity_field, need).token)
        if grants:
            return Q("terms", **{self.grants_field: grants})
        return None


class Creator(EntityNeedsGenerator):
    """Allows the creator of the request."""

    entity_field = 'created_by'


class Receiver(EntityNeedsGenerator):
    """Allows the receiver of the request."""

    entity_field = 'receiver'


class Commenter(Generator):
    """The user who created a specific comment."""

    def needs(self, event=None, request=None, **kwargs):
        """Enabling Needs."""
        if event.created_by is not None:
            return event.created_by.get_needs()
        return []

    def query_filter(self, identity=None, **kwargs):
        """Filters for current identity as creator."""
        raise RuntimeError('The generator cannot be used for searching.')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Custom exceptions used in the Invenio-Requests module."""


class ActionError(Exception):
    """Exception indicating an error related to the action."""

    def __init__(self, action, reason):
        """Constructor.

        :param action: The name of the action in question.
        :param reason: Description of what went wrong.
        """
        self.action = action
        self.reason = reason

    def __str__(self):
        """Return str(self)."""
        return f"Error for action '{self.action}': {self.reason}"


class NoSuchActionError(ActionError):
    """There was no such action available."""

    def __init__(self, action, reason=None):
        """Constructor.

        :param action: The name of the action.
        """
        reason = reason or "No such action available"
        super().__init__(action, reason)


class CannotExecuteActionError(ActionError):
    """The selected action could not be executed."""

    def __init__(self, action, reason=None):
        """Constructor.

        :param action: The name of the action.
        """
        reason = reason or "Could not execute the action"
        super().__init__(action, reason)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Component for creating request numbers."""

from invenio_records_resources.services.records.components import (
    DataComponent,
    ServiceComponent,
)


class RequestNumberComponent(ServiceComponent):
    """Component for assigning request numbers to new requests."""

    def create(self, identity, data=None, record=None, **kwargs):
        """Create identifier when record is created."""
        type(record).number.assign(record)


class EntityReferencesComponent(ServiceComponent):
    """Component for initializing a request's entity references."""

    def create(self, identity, data=None, record=None, **kwargs):
        """Initialize the entity reference fields of a request."""
        for field in ("created_by", "receiver", "topic"):
            if field in kwargs:
                setattr(record, field, kwargs[field])


class RequestDataComponent(DataComponent):
    """Request variant of DataComponent using dynamic schema."""

    def update(self, identity, data=None, record=None, **kwargs):
        """Update an existing record (request)."""
        if record.status == 'created':
            keys = ('title', 'description', 'payload', 'receiver', 'topic')
        else:
            keys = ('title', 'description')

        for k in keys:
            if k in data:
                record[k] = data[k]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
# Copyright (C) 2021-2022 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Requests resource."""


from flask import g
from flask_resources import resource_requestctx, response_handler, route
from invenio_records_resources.resources import RecordResource
from invenio_records_resources.resources.records.resource import (
    request_data,
    request_extra_args,
    request_headers,
    request_search_args,
    request_view_args,
)
from invenio_records_resources.resources.records.utils import es_preference


#
# Resource
#
class RequestsResource(RecordResource):
    """Resource for generic requests."""

    def create_url_rules(self):
        """Create the URL rules for the record resource."""
        routes = self.config.routes
        return [
            route("GET", routes["list"], self.search),
            route("GET", routes["item"], self.read),
            route("PUT", routes["item"], self.update),
            route("DELETE", routes["item"], self.delete),
            route("POST", routes["action"], self.execute_action),
        ]

    @request_extra_args
    @request_search_args
    @request_view_args
    @response_handler(many=True)
    def search(self):
        """Perform a search over the items."""
        hits = self.service.search(
            identity=g.identity,
            params=resource_requestctx.args,
            es_preference=es_preference(),
            expand=resource_requestctx.args["expand"],
        )
        return hits.to_dict(), 200

    @request_extra_args
    @request_view_args
    @response_handler()
    def read(self):
        """Read an item."""
        item = self.service.read(
            id_=resource_requestctx.view_args["id"],
            identity=g.identity,
            expand=resource_requestctx.args["expand"],
        )
        return item.to_dict(), 200

    @request_extra_args
    @request_headers
    @request_view_args
    @request_data
    @response_handler()
    def update(self):
        """Update an item."""
        # TODO should we allow updating of requests in this general resource?
        item = self.service.update(
            id_=resource_requestctx.view_args["id"],
            identity=g.identity,
            data=resource_requestctx.data,
            expand=resource_requestctx.args["expand"],
        )
        return item.to_dict(), 200

    @request_headers
    @request_view_args
    def delete(self):
        """Delete an item."""
        self.service.delete(
            id_=resource_requestctx.view_args["id"],
            identity=g.identity,
        )
        return "", 204

    @request_extra_args
    @request_view_args
    @request_headers
    @request_data
    @response_handler()
    def execute_action(self):
        """Execute action."""
        item = self.service.execute_action(
            identity=g.identity,
            id_=resource_requestctx.view_args["id"],
            action=resource_requestctx.view_args["action"],
            data=resource_requestctx.data,
            expand=resource_requestctx.args["expand"],
        )
        return item.to_dict(), 200



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Request ui views module."""

from flask import Blueprint, current_app, render_template
from flask_login import current_user
from invenio_pidstore.errors import PIDDeletedError, PIDDoesNotExistError
from invenio_records_resources.services.errors import PermissionDeniedError

#
# Error handlers
#


def not_found_error(error):
    """Handler for 'Not Found' errors."""
    return render_template(current_app.config['THEME_404_TEMPLATE']), 404


def record_tombstone_error(error):
    """Tombstone page."""
    return render_template("invenio_requests/tombstone.html"), 410


def record_permission_denied_error(error):
    """Handle permission denier error on record views."""
    if not current_user.is_authenticated:
        # trigger the flask-login unauthorized handler
        return current_app.login_manager.unauthorized()
    return render_template(current_app.config['THEME_403_TEMPLATE']), 403


def create_ui_blueprint(app):
    """Register blueprint routes on app."""
    routes = app.config.get("REQUESTS_ROUTES")

    blueprint = Blueprint(
        "invenio_requests",
        __name__,
        template_folder="../templates",
        static_folder='../static'
    )

    # Register error handlers
    blueprint.register_error_handler(
        PermissionDeniedError,
        record_permission_denied_error,
    )
    blueprint.register_error_handler(
        PIDDeletedError,
        record_tombstone_error,
    )
    blueprint.register_error_handler(PIDDoesNotExistError, not_found_error)

    return blueprint



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
# Copyright (C) 2021-2022 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for generic and customizable requests."""

from .customizations import RequestAction
from .ext import InvenioRequests
from .proxies import (
    current_event_type_registry,
    current_events_service,
    current_request_type_registry,
    current_requests,
    current_requests_resource,
    current_requests_service,
)

__version__ = "0.3.14"

__all__ = (
    "__version__",
    "current_event_type_registry",
    "current_events_service",
    "current_request_type_registry",
    "current_requests_resource",
    "current_requests_service",
    "current_requests",
    "InvenioRequests",
    "RequestAction",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""JS/CSS bundles for communities.

You include one of the bundles in a page like the example below (using
``comunities`` bundle as an example):

.. code-block:: html

    {{ webpack['communities.js']}}

"""

from invenio_assets.webpack import WebpackThemeBundle

requests = WebpackThemeBundle(
    __name__,
    'assets',
    default='semantic-ui',
    themes={
        'semantic-ui': dict(
            entry={
                'invenio-requests-theme':
                    './less/invenio_requests/theme.less',
                'invenio-requests-base':
                    './js/invenio_requests/requestsAppInit.js',
            },
            dependencies={
                'semantic-ui-css': '^2.4.1',
                'semantic-ui-react': '^2.1.1',
                '@ckeditor/ckeditor5-build-classic': '^16.0.0',
                '@ckeditor/ckeditor5-react': '^2.1.0',
                'axios': '^0.19.0',
                'lodash': '^4.17.15',
                'luxon': '^1.21.1',
                'path': '^0.12.7',
                'prop-types': '^15.7.2',
                'qs': '^6.9.1',
                'react': '^16.12.0',
                'react-dom': '^16.11.0',
                'yup': '^0.27.0',
                'react-overridable': '^0.0.3',
                '@semantic-ui-react/css-patch': '^1.0.0',
                "redux": "^4.0.5",
                "redux-devtools-extension": "^2.13.8",
                "redux-thunk": "^2.3.0",
                "react-redux": "^7.2.0",
                "i18next": "^20.3.0",
                "i18next-browser-languagedetector": "^6.1.0",
                "react-i18next": "^11.11.0",
            },
            aliases={
                '@translations/invenio_requests':
                    'translations/invenio_requests',
                '@js/invenio_requests':
                    'js/invenio_requests'
            }
        ),
    }
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Request Events Service Config."""

from invenio_records_resources.services import (
    Link,
    RecordServiceConfig,
    ServiceSchemaWrapper,
)
from invenio_records_resources.services.records.components import DataComponent
from invenio_records_resources.services.records.links import pagination_links
from invenio_records_resources.services.records.results import RecordItem, RecordList

from ...records.api import Request, RequestEvent
from ..configurator import ConfiguratorMixin, FromConfig
from ..permissions import PermissionPolicy
from ..requests.components import EntityReferencesComponent
from ..schemas import RequestEventSchema


class RequestEventItem(RecordItem):
    """RequestEvent result item."""

    @property
    def id(self):
        """Id property."""
        return self._record.id


class RequestEventList(RecordList):
    """RequestEvent result item."""

    @property
    def hits(self):
        """Iterator over the hits."""
        for hit in self._results:
            # Load dump
            record = self._service.record_cls.loads(hit.to_dict())

            # Project the record
            schema = ServiceSchemaWrapper(
                self._service, record.type.marshmallow_schema())
            projection = schema.dump(
                record,
                context=dict(
                    identity=self._identity,
                    record=record,
                )
            )

            if self._links_item_tpl:
                projection['links'] = self._links_item_tpl.expand(record)

            yield projection


class RequestEventLink(Link):
    """Link variables setter for RequestEvent links."""

    @staticmethod
    def vars(record, vars):
        """Variables for the URI template."""
        vars.update({"id": record.id, "request_id": record.request_id})


class RequestEventsServiceConfig(RecordServiceConfig, ConfiguratorMixin):
    """Config."""

    request_cls = Request
    permission_policy_cls = FromConfig(
        "REQUESTS_PERMISSION_POLICY", default=PermissionPolicy
    )
    schema = RequestEventSchema
    record_cls = RequestEvent
    result_item_cls = RequestEventItem
    result_list_cls = RequestEventList

    # ResultItem configurations
    links_item = {
        "self": RequestEventLink("{+api}/requests/{request_id}/comments/{id}"),
    }
    links_search = pagination_links("{+api}/requests/{request_id}/timeline{?args*}")
